"""
test_physics.py - tests for physics.py

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com

GNU General Public License v3 (GNU GPLv3)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from numpy.testing import assert_allclose

from amoeba2 import physics


def test_calc_thermal_fwhm():
    fwhm = physics.calc_thermal_fwhm(100.0).eval()
    assert not np.isnan(fwhm)


def test_calc_nonthermal_fwhm():
    fwhm = physics.calc_nonthermal_fwhm(1.0, 10.0, 0.5)
    assert not np.isnan(fwhm)


def test_calc_Tex():
    Tex = physics.calc_Tex(1800.0, 0.1)
    assert not np.isnan(Tex)


def test_calc_line_profile():
    velo_axis = np.linspace(-10.0, 10.0, 1001)
    velocity = np.array([0.0, 1.0])
    fwhm = np.array([1.0, 2.0])
    line_profile = physics.calc_line_profile(velo_axis, velocity, fwhm).eval()
    assert line_profile.shape == (1001, 2)
    assert_allclose(line_profile.sum(axis=0) * (velo_axis[1] - velo_axis[0]), np.ones(2))
    exp_line_profile = np.array(
        [
            np.sqrt(4.0 * np.log(2.0) / np.pi),
            0.5 * np.sqrt(np.log(2.0) / np.pi),
        ]
    )
    assert_allclose(line_profile[500, :], exp_line_profile)


def test_calc_optical_depth():
    velo_axis = np.linspace(-10.0, 10.0, 1001)
    velocity = np.array([0.0, 2.0])
    fwhm = np.array([1.0, 3.0])
    line_profile = physics.calc_line_profile(velo_axis, velocity, fwhm).eval()

    gu = 3
    gl = 5
    Nl = np.array([1.0e11, 1.0e13])
    boltz_factor = np.exp(-np.array([-0.1, 0.1]))
    freq = 1800.0
    Aul = 1.0e-13
    optical_depth = physics.calc_optical_depth(gu, gl, Nl, boltz_factor, line_profile, freq, Aul)
    assert optical_depth.shape == (1001, 2)
    assert np.all(optical_depth[:, 0] < 0)
    assert np.all(optical_depth[:, 1] > 0)


def test_rj_temperature():
    Tex = np.array([-10.0, 10.0])
    rj_temp = physics.rj_temperature(1800.0, Tex).eval()
    assert rj_temp[0] < rj_temp[1]


def test_radiative_transfer():
    velo_axis = np.linspace(-10.0, 10.0, 1001)
    velocity = np.array([0.0, 2.0])
    fwhm = np.array([1.0, 3.0])
    line_profile = physics.calc_line_profile(velo_axis, velocity, fwhm).eval()

    gu = 3
    gl = 5
    Nu = np.array([1.0e12, 1.0e12])
    Nl = np.array([1.0e11, 1.0e13])
    freq = 1800.0
    Aul = 1.0e-13
    optical_depth = physics.calc_optical_depth(gu, gl, Nu, Nl, line_profile, freq, Aul)

    Tex = np.array([-10.0, 10.0])
    bg_temp = 2.7
    tb = physics.radiative_transfer(freq, optical_depth, Tex, bg_temp).eval()
    assert tb.shape == (1001,)



"""emission_absorption_model.py
EmissionAbsorptionModel definition

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Iterable, Optional

import pymc as pm
import pytensor.tensor as pt
import numpy as np

from bayes_spec import BaseModel

from amoeba2.utils import _G, _OH, get_molecule_data
from amoeba2 import physics


class EmissionAbsorptionModel(BaseModel):
    """Definition of the TBTauModel. SpecData keys must be "absorption_1612", "absorption_1665",
    "absorption_1667", "absorption_1720", "emission_1612", "emission_1665", "emission_1667",
    and "emission_1720".
    """

    def __init__(self, *args, mol_data: Optional[dict] = None, bg_temp: float = 3.77, **kwargs):
        """Initialize a new EmissionAbsorptionModel instance.

        Parameters
        ----------
        mol_data : Optional[dict], optional
            OH molecular data dictionary returned by get_molecule_data(). If None, it will
            be downloaded. Default is None.
        bg_temp : float, optional
            Assumed background temperature (K), by default 3.77
        """
        # Initialize BaseModel
        super().__init__(*args, **kwargs)

        # get OH data
        if mol_data is None:
            self.mol_data = get_molecule_data()
        else:
            self.mol_data = mol_data

        # Save background temperature
        self.bg_temp = bg_temp

        # Add states, transitions to model
        self.model.add_coords(
            {
                "state": [0, 1, 2, 3],
                "transition": ["1612", "1665", "1667", "1720"],
            }
        )

        # Select features used for posterior clustering
        self._cluster_features += [
            "log10_N0",
            "velocity",
            "fwhm",
        ]

        # Define TeX representation of each parameter
        self.var_name_map.update(
            {
                "log10_N0": r"$\log_{10} N_0$ (cm$^{-2}$)",
                "log_boltz_factor": r"$\ln B$",
                "log10_depth": r"log$_{10}$ $d$ (pc)",
                "log10_Tkin": r"$\log_{10} T_{\rm kin}$ (K)",
                "Tex": r"$T_{\rm ex}$ (K)",
                "velocity": r"$V_{\rm LSR}$ (km s$^{-1}$)",
                "log10_nth_fwhm_1pc": r"log$_{10}$ $\Delta V_{\rm 1 pc}$ (km s$^{-1}$)",
                "depth_nth_fwhm_power": r"$\alpha$",
                "log10_N": r"$\log_{10} N$ (cm$^{-2}$)",
                "fwhm_thermal": r"$\Delta V_{\rm th}$ (km s$^{-1}$)",
                "fwhm_nonthermal": r"$\Delta V_{\rm nth}$ (km s$^{-1}$)",
                "fwhm": r"$\Delta V$ (km s$^{-1}$)",
                "rms_absorption": r"rms$_{\tau}$",
            }
        )

    def add_priors(
        self,
        prior_log10_N0: Iterable[float] = [13.0, 1.0],
        prior_log_boltz_factor: Iterable[float] = [-0.1, 0.1],
        prior_log10_depth: Iterable[float] = [0.0, 0.25],
        prior_log10_Tkin: Iterable[float] = [2.0, 1.0],
        prior_velocity: Iterable[float] = [0.0, 10.0],
        prior_log10_nth_fwhm_1pc: Iterable[float] = [0.2, 0.1],
        prior_depth_nth_fwhm_power: Iterable[float] = [0.4, 0.1],
        prior_baseline_coeffs: Optional[dict[str, Iterable[float]]] = None,
        ordered: bool = False,
        mainline_pos_tau: bool = False,
    ):
        """Add priors and deterministics to the model.

        Parameters
        ----------
        prior_log10_N0 : Iterable[float], optional
            Prior distribution on log10 column density (cm-2) in lowest energy state, by default [13.0, 1.0], where
            log10_N0 ~ Normal(mu=prior[0], sigma=prior[1])
        prior_log_boltz_factor : Iterable[float], optional
            Prior distribution on log Boltzmann factor = -h*freq/(k*Tex), by default [-0.1, 0.1], where
            log_boltz_factor ~ Normal(mu=prior[0], sigma=prior[1])
        prior_log10_depth : Iterable[float], optional
            Prior distribution on log10 depth (pc), by default [0.0, 0.25], where
            log10_depth ~ Normal(mu=prior[0], sigma=prior[1])
        prior_log10_Tkin : Iterable[float], optional
            Prior distribution on log10 kinetic temperature (K), by default [2.0, 1.0], where
            log10_Tkin ~ Normal(mu=prior[0], sigma=prior[1])
        prior_velocity : Iterable[float], optional
            Prior distribution on centroid velocity (km s-1), by default [0.0, 10.0], where
            velocity ~ Normal(mu=prior[0], sigma=prior[1])
        prior_log10_nth_fwhm_1pc : Iterable[float], optional
            Prior distribution on non-thermal line width at 1 pc, by default [0.2, 0.1], where
            log10_nth_fwhm_1pc ~ Normal(mu=prior[0], sigma=prior[1])
        prior_depth_nth_fwhm_power : Iterable[float], optional
            Prior distribution on depth vs. non-thermal line width power law index, by default [0.4, 0.1], where
            depth_nth_fwhm_power ~ Normal(mu=prior[0], sigma=prior[1])
        prior_baseline_coeffs : Optional[dict[str, Iterable[float]]], optional
            Width of normal prior distribution on the normalized baseline polynomial coefficients.
            Keys are dataset names and values are lists of length `baseline_degree+1`. If None, use
            `[1.0]*(baseline_degree+1)` for each dataset, by default None
        ordered : bool, optional
            If True, assume ordered velocities (optically thin assumption), by default False. If True, the prior
            distribution on the velocity becomes
            velocity(cloud = n) ~ prior[0] + sum_i(velocity[i < n]) + Gamma(alpha=2.0, beta=1.0/prior[1])
        mainline_pos_tau: bool, optional
            If True, assume positive main line excitation temperatures, by default False. If True, the
            prior distribution on the Boltzmann factor for the main lines (1665 and 1667) becomes
            log_boltz_factor ~ HalfNormal(sigma=prior[1])
        """
        # add polynomial baseline priors
        super().add_baseline_priors(prior_baseline_coeffs=prior_baseline_coeffs)

        with self.model:
            # state 0 column density (cm-2; shape: clouds)
            log10_N0_norm = pm.Normal("log10_N0_norm", mu=0.0, sigma=1.0, dims="cloud")
            log10_N0 = pm.Deterministic("log10_N0", prior_log10_N0[0] + prior_log10_N0[1] * log10_N0_norm, dims="cloud")

            # 2 -> 1 (1612 MHz) Boltzmann factor (shape: clouds)
            log_boltz_factor_1612_norm = pm.Normal("log_boltz_factor_1612_norm", mu=0.0, sigma=1.0, dims="cloud")
            log_boltz_factor_1612 = pm.Deterministic(
                "log_boltz_factor_1612",
                prior_log_boltz_factor[0] + prior_log_boltz_factor[1] * log_boltz_factor_1612_norm,
                dims="cloud",
            )

            if mainline_pos_tau:
                # 0 -> 2 (1665 MHz) Boltzmann factor must be positive (shape: clouds)
                log_boltz_factor_1665_norm = pm.HalfNormal("log_boltz_factor_1665_norm", sigma=1.0, dims="cloud")
                log_boltz_factor_1665 = pm.Deterministic(
                    "log_boltz_factor_1665", -prior_log_boltz_factor[1] * log_boltz_factor_1665_norm, dims="cloud"
                )

                # 1 -> 3 (1667 MHz) Boltzmann factor must be positive (shape: clouds)
                log_boltz_factor_1667_norm = pm.HalfNormal("log_boltz_factor_1667_norm", sigma=1.0, dims="cloud")
                log_boltz_factor_1667 = pm.Deterministic(
                    "log_boltz_factor_1667", -prior_log_boltz_factor[1] * log_boltz_factor_1667_norm, dims="cloud"
                )
            else:
                # 0 -> 2 (1665 MHz) Boltzmann factor (shape: clouds)
                log_boltz_factor_1665_norm = pm.Normal("log_boltz_factor_1665_norm", mu=0.0, sigma=1.0, dims="cloud")
                log_boltz_factor_1665 = pm.Deterministic(
                    "log_boltz_factor_1665",
                    prior_log_boltz_factor[0] + prior_log_boltz_factor[1] * log_boltz_factor_1665_norm,
                    dims="cloud",
                )

                # 1 -> 3 (1667 MHz) Boltzmann factor (shape: clouds)
                log_boltz_factor_1667_norm = pm.Normal("log_boltz_factor_1667_norm", mu=0.0, sigma=1.0, dims="cloud")
                log_boltz_factor_1667 = pm.Deterministic(
                    "log_boltz_factor_1667",
                    prior_log_boltz_factor[0] + prior_log_boltz_factor[1] * log_boltz_factor_1667_norm,
                    dims="cloud",
                )

            # 3 -> 0 (1720 MHz) Boltzmann factor (shape: clouds)
            # Boltzmann factor sum rule
            _ = pm.Deterministic(
                "log_boltz_factor_1720", log_boltz_factor_1665 + log_boltz_factor_1667 - log_boltz_factor_1612
            )

            # State column density (cm-2; shape: state, cloud)
            log10_N2 = pm.Deterministic(
                "log10_N2", log10_N0 + log_boltz_factor_1665 * np.log10(np.e) - pt.log10(_G[0] / _G[2]), dims="cloud"
            )
            log10_N1 = pm.Deterministic(
                "log10_N1", log10_N2 - log_boltz_factor_1612 * np.log10(np.e) + pt.log10(_G[1] / _G[2]), dims="cloud"
            )
            _ = pm.Deterministic(
                "log10_N3", log10_N1 + log_boltz_factor_1667 * np.log10(np.e) - pt.log10(_G[1] / _G[3]), dims="cloud"
            )

            # depth (pc; shape: clouds)
            log10_depth_norm = pm.Normal("log10_depth_norm", mu=0.0, sigma=1.0, dims="cloud")
            log10_depth = pm.Deterministic(
                "log10_depth",
                prior_log10_depth[0] + prior_log10_depth[1] * log10_depth_norm,
                dims="cloud",
            )

            # kinetic temperature (K; shape: clouds)
            log10_Tkin_norm = pm.Normal("log10_Tkin_norm", mu=0.0, sigma=1.0, dims="cloud")
            log10_Tkin = pm.Deterministic(
                "log10_Tkin",
                prior_log10_Tkin[0] + prior_log10_Tkin[1] * log10_Tkin_norm,
                dims="cloud",
            )

            # Center velocity (km s-1; shape: clouds)
            if ordered:
                velocity_offset_norm = pm.Gamma("velocity_norm", alpha=2.0, beta=1.0, dims="cloud")
                velocity_offset = velocity_offset_norm * prior_velocity[1]
                _ = pm.Deterministic(
                    "velocity",
                    prior_velocity[0] + pm.math.cumsum(velocity_offset),
                    dims="cloud",
                )
            else:
                velocity_norm = pm.Normal(
                    "velocity_norm",
                    mu=0.0,
                    sigma=1.0,
                    dims="cloud",
                )
                _ = pm.Deterministic(
                    "velocity",
                    prior_velocity[0] + prior_velocity[1] * velocity_norm,
                    dims="cloud",
                )

            # Non-thermal FWHM at 1 pc (km s-1; shape: clouds)
            log10_nth_fwhm_1pc_norm = pm.Normal("log10_nth_fwhm_1pc_norm", mu=0.0, sigma=1.0)
            log10_nth_fwhm_1pc = pm.Deterministic(
                "log10_nth_fwhm_1pc",
                prior_log10_nth_fwhm_1pc[0] + prior_log10_nth_fwhm_1pc[1] * log10_nth_fwhm_1pc_norm,
            )

            # Non-thermal FWHM vs. depth power law index (shape: clouds)
            depth_nth_fwhm_power_norm = pm.Normal("depth_nth_fwhm_power_norm", mu=0.0, sigma=1.0)
            depth_nth_fwhm_power = pm.Deterministic(
                "depth_nth_fwhm_power",
                prior_depth_nth_fwhm_power[0] + prior_depth_nth_fwhm_power[1] * depth_nth_fwhm_power_norm,
            )

            # Thermal FWHM (km/s; shape: clouds)
            fwhm_thermal = pm.Deterministic("fwhm_thermal", physics.calc_thermal_fwhm(10.0**log10_Tkin), dims="cloud")

            # Non-thermal FWHM (km/s; shape: clouds)
            fwhm_nonthermal = pm.Deterministic(
                "fwhm_nonthermal",
                physics.calc_nonthermal_fwhm(10.0**log10_depth, 10.0**log10_nth_fwhm_1pc, depth_nth_fwhm_power),
                dims="cloud",
            )

            # FWHM (km/s; shape: clouds)
            _ = pm.Deterministic("fwhm", pt.sqrt(fwhm_thermal**2.0 + fwhm_nonthermal**2.0), dims="cloud")

            # Excitation temperatures (K; shape: clouds)
            for freq, label in zip(self.mol_data["freq"], self.model.coords["transition"]):
                _ = pm.Deterministic(
                    f"Tex_{label}",
                    physics.calc_Tex(freq, self.model[f"log_boltz_factor_{label}"]),
                    dims="cloud",
                )

    def predict_emission_absorption(self) -> dict:
        """Predict the emission and absorption spectra from the model parameters.

        Returns
        -------
        dict
            Emission spectra (brightness temp; K) for 1612, 1665, 1667, and 1720 MHz transitions
        dict
            Absorption spectra (1 - exp(-tau)) for 1612, 1665, 1667, and 1720 MHz transitions
        """
        emission = {}
        absorption = {}
        for freq, Aul, label in zip(self.mol_data["freq"], self.mol_data["Aul"], self.model.coords["transition"]):
            # Line profile (km-1 s; shape: spectral, cloud)
            line_profile_absorption = physics.calc_line_profile(
                self.data[f"absorption_{label}"].spectral,
                self.model["velocity"],
                self.model["fwhm"],
            )
            line_profile_emission = physics.calc_line_profile(
                self.data[f"emission_{label}"].spectral,
                self.model["velocity"],
                self.model["fwhm"],
            )

            # Optical depth spectrum (shape: spectral, cloud)
            tau_spectrum_absorption = physics.calc_optical_depth(
                _G[_OH[label][0]],
                _G[_OH[label][1]],
                10.0 ** self.model[f"log10_N{_OH[label][1]}"],
                pt.exp(self.model[f"log_boltz_factor_{label}"]),
                line_profile_absorption,
                freq,
                Aul,
            )
            tau_spectrum_emission = physics.calc_optical_depth(
                _G[_OH[label][0]],
                _G[_OH[label][1]],
                10.0 ** self.model[f"log10_N{_OH[label][1]}"],
                pt.exp(self.model[f"log_boltz_factor_{label}"]),
                line_profile_emission,
                freq,
                Aul,
            )

            # Sum over clouds
            absorption[label] = 1.0 - pt.exp(-tau_spectrum_absorption.sum(axis=1))

            # Radiative transfer
            emission[label] = physics.radiative_transfer(
                freq,
                tau_spectrum_emission,
                self.model[f"Tex_{label}"],
                self.bg_temp,
            )
        return emission, absorption

    def add_likelihood(self):
        """Add likelihood to the model."""
        # Predict emission and absorption spectra
        emission, absorption = self.predict_emission_absorption()

        # Get baseline models
        baseline_models = self.predict_baseline()

        with self.model:
            for label in self.model.coords["transition"]:
                _ = pm.Normal(
                    f"absorption_{label}",
                    mu=absorption[label] + baseline_models[f"absorption_{label}"],
                    sigma=self.data[f"absorption_{label}"].noise,
                    observed=self.data[f"absorption_{label}"].brightness,
                )
                _ = pm.Normal(
                    f"emission_{label}",
                    mu=emission[label] + baseline_models[f"emission_{label}"],
                    sigma=self.data[f"emission_{label}"].noise,
                    observed=self.data[f"emission_{label}"].brightness,
                )



"""
test_utils.py - tests for utils.py

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com

GNU General Public License v3 (GNU GPLv3)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import pytest
from amoeba2 import utils


def test_get_molecule_data():
    with pytest.raises(ValueError):
        utils.get_molecule_data(molecule="BAD MOLECULE NAME")
    data = utils.get_molecule_data()
    assert "freq" in data.keys()
    assert "Aul" in data.keys()
    assert "degu" in data.keys()
    assert "Eu" in data.keys()
    assert "relative_int" in data.keys()
    assert "log10_Q_terms" in data.keys()



"""absorption_model.py
AbsorptionModel definition

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Iterable, Optional

import pymc as pm
import pytensor.tensor as pt

from bayes_spec import BaseModel

from amoeba2.utils import get_molecule_data
from amoeba2 import physics


class AbsorptionModel(BaseModel):
    """Definition of the AbsorptionModel. SpecData keys must be "absorption_1612", "absorption_1665",
    "absorption_1667", and "absorption_1720"."""

    def __init__(self, *args, mol_data: Optional[dict] = None, **kwargs):
        """Initialize a new AbsorptionModel instance.

        Parameters
        ----------
        mol_data : Optional[dict], optional
            OH molecular data dictionary returned by get_molecule_data(). If None, it will
            be downloaded. Default is None.
        """
        # Initialize BaseModel
        super().__init__(*args, **kwargs)

        # get OH data
        if mol_data is None:
            self.mol_data = get_molecule_data()
        else:
            self.mol_data = mol_data

        # Select features used for posterior clustering
        self._cluster_features += [
            "tau_1612",
            "tau_1665",
            "tau_1667",
            "velocity",
            "fwhm",
        ]

        # Add transitions to model
        self.model.add_coords(
            {
                "transition": ["1612", "1665", "1667", "1720"],
            }
        )

        # Define TeX representation of each parameter
        self.var_name_map.update(
            {
                "tau_1612": r"$\tau_{1612}$",
                "tau_1665": r"$\tau_{1665}$",
                "tau_1667": r"$\tau_{1667}$",
                "tau_1720": r"$\tau_{1720}$",
                "log10_depth": r"log$_{10}$ $d$ (pc)",
                "log10_Tkin": r"$\log_{10} T_{\rm kin}$ (K)",
                "velocity": r"$V_{\rm LSR}$ (km s$^{-1}$)",
                "log10_nth_fwhm_1pc": r"log$_{10}$ $\Delta V_{\rm 1 pc}$ (km s$^{-1}$)",
                "depth_nth_fwhm_power": r"$\alpha$",
                "fwhm_thermal": r"$\Delta V_{\rm th}$ (km s$^{-1}$)",
                "fwhm_nonthermal": r"$\Delta V_{\rm nth}$ (km s$^{-1}$)",
                "fwhm": r"$\Delta V$ (km s$^{-1}$)",
            }
        )

    def add_priors(
        self,
        prior_tau: Iterable[float] = [0.1, 0.1],
        prior_log10_depth: Iterable[float] = [0.0, 0.25],
        prior_log10_Tkin: Iterable[float] = [2.0, 1.0],
        prior_velocity: Iterable[float] = [0.0, 10.0],
        prior_log10_nth_fwhm_1pc: Iterable[float] = [0.2, 0.1],
        prior_depth_nth_fwhm_power: Iterable[float] = [0.4, 0.1],
        prior_baseline_coeffs: Optional[dict[str, Iterable[float]]] = None,
        ordered: bool = False,
        mainline_pos_tau: bool = False,
    ):
        """Add priors and deterministics to the model.

        Parameters
        ----------
        prior_tau : float, optional
            Prior distribution on optical depth, by default [0.1, 0.1], where
            tau ~ Normal(mu=prior[0], sigma=prior[1])
        prior_log10_depth : Iterable[float], optional
            Prior distribution on log10 depth (pc), by default [0.0, 0.25], where
            log10_depth ~ Normal(mu=prior[0], sigma=prior[1])
        prior_log10_Tkin : Iterable[float], optional
            Prior distribution on log10 kinetic temperature (K), by default [2.0, 1.0], where
            log10_Tkin ~ Normal(mu=prior[0], sigma=prior[1])
        prior_velocity : Iterable[float], optional
            Prior distribution on centroid velocity (km s-1), by default [0.0, 10.0], where
            velocity ~ Normal(mu=prior[0], sigma=prior[1])
        prior_log10_nth_fwhm_1pc : Iterable[float], optional
            Prior distribution on non-thermal line width at 1 pc, by default [0.2, 0.1], where
            log10_nth_fwhm_1pc ~ Normal(mu=prior[0], sigma=prior[1])
        prior_depth_nth_fwhm_power : Iterable[float], optional
            Prior distribution on depth vs. non-thermal line width power law index, by default [0.4, 0.1], where
            depth_nth_fwhm_power ~ Normal(mu=prior[0], sigma=prior[1])
        prior_baseline_coeffs : Optional[dict[str, Iterable[float]]], optional
            Width of normal prior distribution on the normalized baseline polynomial coefficients.
            Keys are dataset names and values are lists of length `baseline_degree+1`. If None, use
            `[1.0]*(baseline_degree+1)` for each dataset, by default None
        ordered : bool, optional
            If True, assume ordered velocities (optically thin assumption), by default False. If True, the prior
            distribution on the velocity becomes
            velocity(cloud = n) ~ prior[0] + sum_i(velocity[i < n]) + Gamma(alpha=2.0, beta=1.0/prior[1])
        mainline_pos_tau: bool, optional
            If True, assume positive main line optical depths, by default False. If True, the
            prior distribution on the main line (1665 and 1667 MHz) optical depths becomes
            tau ~ HalfNormal(sigma=prior[1])
        """
        # add polynomial baseline priors
        super().add_baseline_priors(prior_baseline_coeffs=prior_baseline_coeffs)

        with self.model:
            # optical depth
            tau_1612_norm = pm.Normal("tau_1612_norm", mu=0.0, sigma=1.0, dims="cloud")
            tau_1612 = pm.Deterministic("tau_1612", prior_tau[0] + tau_1612_norm * prior_tau[1], dims="cloud")

            if mainline_pos_tau:
                tau_1665_norm = pm.HalfNormal("tau_1665_norm", sigma=1.0, dims="cloud")
                tau_1665 = pm.Deterministic("tau_1665", tau_1665_norm * prior_tau[1], dims="cloud")

                tau_1667_norm = pm.HalfNormal("tau_1667_norm", sigma=1.0, dims="cloud")
                tau_1667 = pm.Deterministic("tau_1667", tau_1667_norm * prior_tau[1], dims="cloud")
            else:
                tau_1665_norm = pm.Normal("tau_1665_norm", mu=0.0, sigma=1.0, dims="cloud")
                tau_1665 = pm.Deterministic("tau_1665", prior_tau[0] + tau_1665_norm * prior_tau[1], dims="cloud")

                tau_1667_norm = pm.Normal("tau_1667_norm", mu=0.0, sigma=1.0, dims="cloud")
                tau_1667 = pm.Deterministic("tau_1667", prior_tau[0] + tau_1667_norm * prior_tau[1], dims="cloud")

            # optical depth sum rule
            _ = pm.Deterministic("tau_1720", tau_1665 / 5.0 + tau_1667 / 9.0 - tau_1612, dims="cloud")

            # depth (pc; shape: clouds)
            log10_depth_norm = pm.Normal("log10_depth_norm", mu=0.0, sigma=1.0, dims="cloud")
            log10_depth = pm.Deterministic(
                "log10_depth",
                prior_log10_depth[0] + prior_log10_depth[1] * log10_depth_norm,
                dims="cloud",
            )

            # kinetic temperature (K; shape: clouds)
            log10_Tkin_norm = pm.Normal("log10_Tkin_norm", mu=0.0, sigma=1.0, dims="cloud")
            log10_Tkin = pm.Deterministic(
                "log10_Tkin",
                prior_log10_Tkin[0] + prior_log10_Tkin[1] * log10_Tkin_norm,
                dims="cloud",
            )

            # Center velocity (km s-1; shape: clouds)
            if ordered:
                velocity_offset_norm = pm.Gamma("velocity_norm", alpha=2.0, beta=1.0, dims="cloud")
                velocity_offset = velocity_offset_norm * prior_velocity[1]
                _ = pm.Deterministic(
                    "velocity",
                    prior_velocity[0] + pm.math.cumsum(velocity_offset),
                    dims="cloud",
                )
            else:
                velocity_norm = pm.Normal(
                    "velocity_norm",
                    mu=0.0,
                    sigma=1.0,
                    dims="cloud",
                )
                _ = pm.Deterministic(
                    "velocity",
                    prior_velocity[0] + prior_velocity[1] * velocity_norm,
                    dims="cloud",
                )

            # Non-thermal FWHM at 1 pc (km s-1; shape: clouds)
            log10_nth_fwhm_1pc_norm = pm.Normal("log10_nth_fwhm_1pc_norm", mu=0.0, sigma=1.0)
            log10_nth_fwhm_1pc = pm.Deterministic(
                "log10_nth_fwhm_1pc",
                prior_log10_nth_fwhm_1pc[0] + prior_log10_nth_fwhm_1pc[1] * log10_nth_fwhm_1pc_norm,
            )

            # Non-thermal FWHM vs. depth power law index (shape: clouds)
            depth_nth_fwhm_power_norm = pm.Normal("depth_nth_fwhm_power_norm", mu=0.0, sigma=1.0)
            depth_nth_fwhm_power = pm.Deterministic(
                "depth_nth_fwhm_power",
                prior_depth_nth_fwhm_power[0] + prior_depth_nth_fwhm_power[1] * depth_nth_fwhm_power_norm,
            )

            # Thermal FWHM (km/s; shape: clouds)
            fwhm_thermal = pm.Deterministic("fwhm_thermal", physics.calc_thermal_fwhm(10.0**log10_Tkin), dims="cloud")

            # Non-thermal FWHM (km/s; shape: clouds)
            fwhm_nonthermal = pm.Deterministic(
                "fwhm_nonthermal",
                physics.calc_nonthermal_fwhm(10.0**log10_depth, 10.0**log10_nth_fwhm_1pc, depth_nth_fwhm_power),
                dims="cloud",
            )

            # FWHM (km/s; shape: clouds)
            _ = pm.Deterministic("fwhm", pt.sqrt(fwhm_thermal**2.0 + fwhm_nonthermal**2.0), dims="cloud")

    def predict_absorption(self) -> dict:
        """Predict the absorption spectra from the model parameters.

        Returns
        -------
        dict
            Optical depth spectra for 1612, 1665, 1667, and 1720 MHz transitions
        """
        absorption = {}
        for label in self.model.coords["transition"]:
            # Line profile (km-1 s; shape: spectral, cloud)
            line_profile = physics.calc_line_profile(
                self.data[f"absorption_{label}"].spectral,
                self.model["velocity"],
                self.model["fwhm"],
            )

            # Optical depth spectrum (shape: spectral, cloud)
            tau_spectrum = self.model[f"tau_{label}"] * line_profile

            # Sum over clouds
            absorption[label] = 1.0 - pt.exp(-tau_spectrum.sum(axis=1))
        return absorption

    def add_likelihood(self):
        """Add likelihood to the model."""
        # Predict optical depth spectra
        absorption = self.predict_absorption()

        # Get baseline models
        baseline_models = self.predict_baseline()

        with self.model:
            for label in self.model.coords["transition"]:
                _ = pm.Normal(
                    f"absorption_{label}",
                    mu=absorption[label] + baseline_models[f"absorption_{label}"],
                    sigma=self.data[f"absorption_{label}"].noise,
                    observed=self.data[f"absorption_{label}"].brightness,
                )



"""
test_emission_absorption_model.py - tests for EmissionAbsorptionModel

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com

GNU General Public License v3 (GNU GPLv3)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from bayes_spec import SpecData
from amoeba2 import EmissionAbsorptionModel
from amoeba2.utils import get_molecule_data


def test_emission_absorption_model():
    velocity = np.linspace(-20.0, 20.0, 1000)
    brightness = np.random.randn(1000)
    data = {
        "absorption_1612": SpecData(velocity, brightness, 1.0),
        "absorption_1665": SpecData(velocity, brightness, 1.0),
        "absorption_1667": SpecData(velocity, brightness, 1.0),
        "absorption_1720": SpecData(velocity, brightness, 1.0),
        "emission_1612": SpecData(velocity, brightness, 1.0),
        "emission_1665": SpecData(velocity, brightness, 1.0),
        "emission_1667": SpecData(velocity, brightness, 1.0),
        "emission_1720": SpecData(velocity, brightness, 1.0),
    }
    model = EmissionAbsorptionModel(data, 2, baseline_degree=1)
    assert isinstance(model.mol_data, dict)
    mol_data = get_molecule_data()
    model = EmissionAbsorptionModel(data, 2, baseline_degree=1, mol_data=mol_data)
    assert isinstance(model.mol_data, dict)
    model.add_priors()
    model.add_likelihood()
    assert model._validate()


def test_emission_absorption_model_ordered():
    velocity = np.linspace(-20.0, 20.0, 1000)
    brightness = np.random.randn(1000)
    data = {
        "absorption_1612": SpecData(velocity, brightness, 1.0),
        "absorption_1665": SpecData(velocity, brightness, 1.0),
        "absorption_1667": SpecData(velocity, brightness, 1.0),
        "absorption_1720": SpecData(velocity, brightness, 1.0),
        "emission_1612": SpecData(velocity, brightness, 1.0),
        "emission_1665": SpecData(velocity, brightness, 1.0),
        "emission_1667": SpecData(velocity, brightness, 1.0),
        "emission_1720": SpecData(velocity, brightness, 1.0),
    }
    model = EmissionAbsorptionModel(data, 2, baseline_degree=1)
    assert isinstance(model.mol_data, dict)
    model.add_priors(ordered=True)
    model.add_likelihood()
    assert model._validate()


def test_emission_absorption_model_mainline_pos_tau():
    velocity = np.linspace(-20.0, 20.0, 1000)
    brightness = np.random.randn(1000)
    data = {
        "absorption_1612": SpecData(velocity, brightness, 1.0),
        "absorption_1665": SpecData(velocity, brightness, 1.0),
        "absorption_1667": SpecData(velocity, brightness, 1.0),
        "absorption_1720": SpecData(velocity, brightness, 1.0),
        "emission_1612": SpecData(velocity, brightness, 1.0),
        "emission_1665": SpecData(velocity, brightness, 1.0),
        "emission_1667": SpecData(velocity, brightness, 1.0),
        "emission_1720": SpecData(velocity, brightness, 1.0),
    }
    model = EmissionAbsorptionModel(data, 2, baseline_degree=1)
    assert isinstance(model.mol_data, dict)
    model.add_priors(mainline_pos_tau=True)
    model.add_likelihood()
    assert model._validate()



"""physics.py
Molecular spectroscopy physics utilities

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Iterable

import numpy as np
import pytensor.tensor as pt

import astropy.constants as c

from bayes_spec.utils import gaussian

_K_B = c.k_B.to("erg K-1").value
_H = c.h.to("erg MHz-1").value
_C = c.c.to("km s-1").value
_C_CM_MHZ = c.c.to("cm MHz").value


def calc_thermal_fwhm(kinetic_temp: float, weight: float = 17.0) -> float:
    """Calculate the thermal line broadening assuming a Maxwellian velocity distribution
    (Condon & Ransom eq. 7.35)

    Parameters
    ----------
    kinetic_temp : float
        Kinetic temperature (K)
    weight : float
        Molecular weight (number of protons). By default, 17 (OH molecule)

    Returns
    -------
    float
        Thermal FWHM line width (km s-1)
    """
    # constant = sqrt(8*ln(2)*k_B/m_p)
    const = 0.21394418  # km/s K-1/2
    return const * pt.sqrt(kinetic_temp / weight)


def calc_nonthermal_fwhm(depth: float, nth_fwhm_1pc: float, depth_nth_fwhm_power: float) -> float:
    """Calculate the non-thermal line broadening assuming a power-law size-linewidth relationship.

    Parameters
    ----------
    depth : float
        Line-of-sight depth (pc)
    nth_fwhm_1pc : float
        Non-thermal broadening at 1 pc (km s-1)
    depth_nth_fwhm_power : float
        Power law index

    Returns
    -------
    float
        Non-thermal FWHM line width (km s-1)
    """
    return nth_fwhm_1pc * depth**depth_nth_fwhm_power


def calc_Tex(freq: float, log_boltz_factor: float) -> float:
    """Evaluate the excitation temperature from a given Boltzmann factor.

    Parameters
    ----------
    freq : float
        Frequency (MHz)
    log_boltz_factor : float
        log Boltzmann factor = -h*freq/(k*Tex)

    Returns
    -------
    float
        Excitation temperature
    """
    return -_H * freq / (_K_B * log_boltz_factor)


def calc_line_profile(velo_axis: Iterable[float], velocity: Iterable[float], fwhm: Iterable[float]) -> Iterable[float]:
    """Evaluate the Gaussian line profile, ensuring normalization.

    Parameters
    ----------
    velo_axis : Iterable[float]
        Observed velocity axis (km s-1; length S)
    velocity : Iterable[float]
        Cloud center velocities (km s-1; length C)
    fwhm : Iterable[float]
        Cloud FWHM line widths (km s-1; length C)

    Returns
    -------
    Iterable[float]
        Line profile (km-1 s; shape S x C)
    """
    amp = pt.sqrt(4.0 * pt.log(2.0) / (np.pi * fwhm**2.0))
    profile = gaussian(velo_axis[:, None], amp, velocity, fwhm)

    # normalize
    channel_size = pt.abs(velo_axis[1] - velo_axis[0])
    profile_int = pt.sum(profile, axis=0)
    norm = pt.switch(pt.lt(profile_int, 1.0e-6), 1.0, profile_int * channel_size)
    return profile / norm


def calc_optical_depth(
    gu: int,
    gl: int,
    Nl: Iterable[float],
    boltz_factor: Iterable[float],
    line_profile: Iterable[float],
    freq: float,
    Aul: float,
) -> Iterable[float]:
    """Evaluate the optical depth spectra, from Mangum & Shirley eq. 29

    Parameters
    ----------
    gu : int
        Upper state degeneracy
    gl : int
        Lower state degeneracy
    Nl : Iterable[float]
        Cloud lower state column densities (cm-2; length C)
    boltz_factor : Iterable[float]
        Boltzmann factor = exp(-h*freq/(k*Tex)) (length C)
    line_profile : Iterable[float]
        Line profile (km-1 s; shape S x C)
    freq : float
        Transition frequency (MHz)
    Aul : float
        Transition Einstein A coefficient (s-1)

    Returns
    -------
    Iterable[float]
        Optical depth spectral (shape S x C)
    """
    return (
        _C_CM_MHZ**2.0  # cm2 MHz2
        / (8.0 * np.pi * freq**2.0)  # MHz-2
        * (gu / gl)
        * Aul  # s-1
        * (_C * line_profile / (1e6 * freq))  # Hz-1
        * Nl  # cm-2
        * (1.0 - boltz_factor)
    )


def rj_temperature(freq: float, temp: float) -> float:
    """Calculate the Rayleigh-Jeans equivalent temperature (AKA the brightness temperature)

    Parameters
    ----------
    freq : float
        Frequency (MHz)
    temp : float
        Temperature (K)

    Returns
    -------
    float
        R-J equivalent temperature (K)
    """
    const = _H * freq / _K_B
    return const / (pt.exp(const / temp) - 1.0)


def radiative_transfer(
    freq: float,
    tau: Iterable[float],
    Tex: Iterable[float],
    bg_temp: float,
) -> Iterable[float]:
    """Evaluate the radiative transfer to predict the emission spectrum. The emission
    spectrum is ON - OFF, where ON includes the attenuated emission of the background and
    the clouds, and the OFF is the emission of the background. Order of N clouds is
    assumed to be [nearest, ..., farthest].

    Parameters
    ----------
    freq : float
        Frequency (MHz)
    tau : Iterable[float]
        Optical depth spectra (shape S x N)
    Tex : Iterable[float]
        Cloud excitation temperature (K; length N)
    bg_temp : float
        Assumed background temperature

    Returns
    -------
    Iterable[float]
        Predicted emission brightness temperature spectrum (K; length S)
    """
    front_tau = pt.zeros_like(tau[:, 0:1])
    sum_tau = pt.concatenate([front_tau, pt.cumsum(tau, axis=1)], axis=1)

    # radiative transfer, assuming filling factor = 1.0
    emission_bg = rj_temperature(freq, bg_temp)
    emission_bg_attenuated = emission_bg * pt.exp(-sum_tau[:, -1])
    emission_clouds = rj_temperature(freq, Tex) * (1.0 - pt.exp(-tau))
    emission_clouds_attenuated = emission_clouds * pt.exp(-sum_tau[:, :-1])
    emission = emission_bg_attenuated + emission_clouds_attenuated.sum(axis=1)

    # ON - OFF
    return emission - emission_bg



from setuptools import find_packages, setup
import versioneer

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

setup(
    name="amoeba2",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Automated Molecular Excitation Bayesian line-fitting Algorithm",
    author="Trey V. Wenger",
    author_email="tvwenger@gmail.com",
    packages=find_packages(),
    install_requires=install_requires,
    python_requires=">=3.11",
    license="MIT",
    url="https://github.com/tvwenger/amoeba2",
)



"""
utils.py
Various utilities.

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Optional

import numpy as np
from numpy.polynomial import Polynomial

import astropy.units as u
from astroquery.jplspec import JPLSpec

# OH state degeneracies
_G = [3, 5, 3, 5]

# OH states (upper, lower) associated with transitions
_OH = {
    "1612": [2, 1],
    "1665": [2, 0],
    "1667": [3, 1],
    "1720": [3, 0],
}


def get_molecule_data(
    molecule: Optional[str] = "OH",
    fmin: float = 1.0,
    fmax: float = 2.0,
    rot_state_lower: Optional[int] = 0,
) -> dict:
    """Get molecular transition data from the JPL database. By default fetches
    data for 1612, 1665, 1667, and 1720 MHz OH hyperfine transitions.

    Parameters
    ----------
    molecule : str, optional
        Molecule name, by default "OH"
    fmin : float, optional
        Minimum frequency (GHz), by default 0.0
    fmax : float, optional
        Maximum frequency (GHz), by default 10000.0
    rot_state_lower : Optional[int], optional
        For OH, if not None, then limit to this lower rotational state, which is assumed
        to be the second quantum number returned by JPLSpec, by default 0

    Returns
    -------
    dict
        Molecular transition data, with keys:
        "freq" (Iterable[float]) : Rest frequencies (MHz)
        "Aul" (Iterable[float]) : Einstein A coefficients (s-1)
        "degu" (Iterable[float]) : Upper state degeneracies
        "Eu" (Iterable[float]) : Upper state energies (erg)
        "relative_int" (Iterable[float]) : Relative intensities
        "log10_Q_terms" (Iterable[float]) : Polynomial coefficients for logQ vs. logT (K)

    Raises
    ------
    ValueError
        Molecule not found in JPLSpec database
    """
    # get meta data
    species = JPLSpec.get_species_table()
    if molecule not in species["NAME"]:
        raise ValueError(f"{molecule} not found in database")
    data = species[species["NAME"] == molecule][0]

    # partition function linear fit parameters
    log10_temps = np.log10(data.meta["Temperature (K)"])
    log10_Q = np.array([data[f"QLOG{i}"] for i in range(1, 8)])
    log10_Q_fit = Polynomial.fit(log10_temps, log10_Q, 1).convert()
    log10_Q_terms = log10_Q_fit.coef

    # get transition data
    output = JPLSpec.query_lines(
        min_frequency=fmin * u.GHz,
        max_frequency=fmax * u.GHz,
        molecule=f"{data['TAG']} {molecule}",
    )

    # limit states
    if rot_state_lower is not None:
        output_rot_state = np.array([int(output['QN"'][i][3:5]) for i in range(len(output))])
        good = output_rot_state == rot_state_lower
        output = output[good]

    # rest frequencies, degeneracies, and upper energy levels
    freqs = output["FREQ"].quantity
    freq_MHz = freqs.to(u.MHz).value
    deg = np.array(output["GUP"])
    EL = output["ELO"].quantity.to(u.erg, u.spectral())
    dE = freqs.to(u.erg, u.spectral())
    EU = EL + dE

    # need elower, eupper in inverse centimeter units
    elower_icm = output["ELO"].quantity.to(u.cm**-1).value
    eupper_icm = elower_icm + (freqs.to(u.cm**-1, u.spectral()).value)

    # from Brett McGuire
    # https://github.com/bmcguir2/simulate_lte/blob/1f3f7c666946bc88c8d83c53389556a4c75c2bbd/simulate_lte.py#L2580-L2587

    # LGINT: Base 10 logarithm of the integrated intensity in units of nm2 MHz at 300 K.
    # (See Section 3 for conversions to other units.)
    # see also https://cdms.astro.uni-koeln.de/classic/predictions/description.html#description
    CT = 300.0
    logint = np.array(output["LGINT"])  # this should just be a number
    # from CDMS website
    sijmu = (
        (np.exp(np.float64(-(elower_icm / 0.695) / CT)) - np.exp(np.float64(-(eupper_icm / 0.695) / CT))) ** (-1)
        * ((10**logint) / freq_MHz)
        * (24025.120666)
        * 10.0 ** log10_Q_terms[0]
        * CT ** log10_Q_terms[1]
    )

    # aij formula from CDMS.  Verfied it matched spalatalogue's values
    aij = 1.16395 * 10 ** (-20) * freq_MHz**3 * sijmu / deg

    # relative intensity
    relative_int = 10.0**logint
    relative_int = relative_int / relative_int.sum()

    return {
        "freq": freqs.to(u.MHz).value,  # rest frequency (MHz)
        "Aul": aij,  # Einstein A (s-1)
        "degu": deg,  # upper state degeneracy
        "Eu": EU.to(u.erg).value,  # upper level energy (erg)
        "relative_int": relative_int,  # relative intensity
        "log10_Q_terms": log10_Q_terms,  # partition function linear fit coefficients
    }



__all__ = [
    "AbsorptionModel",
    "EmissionAbsorptionModel",
]

from amoeba2.absorption_model import AbsorptionModel
from amoeba2.emission_absorption_model import EmissionAbsorptionModel

from . import _version

__version__ = _version.get_versions()["version"]
